# -*- coding: utf-8 -*-
from copy import deepcopy
from itertools import zip_longest

from django.conf import settings
from django.dispatch import receiver
from django.test.signals import setting_changed
from django.utils.html import escape

from .finders import CdnFinder


ASSETS = deepcopy(getattr(settings, 'ASSETS_MANAGER_FILES', {}))
SPRITES = deepcopy(list(getattr(settings, 'ASSETS_MANAGER_SPRITES', [])))
USE_TEMPLATES = getattr(settings, 'ASSETS_MANAGER_USE_TEMPLATES', False)


finder = CdnFinder()


def transform_static(path):
	if path.find("static://") != 0:
		return path
	return settings.STATIC_URL + path[9:]


def convert_asset_data(name, asset):
	asset.setdefault("depends", [])

	asset.setdefault("css", [])
	if isinstance(asset["css"], str):
		asset["css"] = [asset["css"]]
	asset['css'] = [transform_static(path) for path in asset['css']]
	asset['css'] = finder.transform_to_cache(name, asset['css'])
	asset['css'] = [escape(item) for item in asset['css']]

	asset.setdefault("js", [])
	if isinstance(asset["js"], str):
		asset["js"] = [asset["js"]]
	asset['js'] = [transform_static(path) for path in asset['js']]
	asset['js'] = finder.transform_to_cache(name, asset['js'])
	asset['js'] = [escape(item) for item in asset['js']]

	asset["css"] = list(zip_longest(asset['css'], {}, fillvalue=''))
	asset["js"] = list(zip_longest(asset['js'], {}, fillvalue=''))
	return asset


@receiver(setting_changed)
def update_settings(**kwargs):
	setting = kwargs.get('setting')
	if setting is not None and setting not in {'ASSETS_MANAGER_FILES', 'ASSETS_MANAGER_SPRITES'}:
		return
	sprites = deepcopy(list(getattr(settings, 'ASSETS_MANAGER_SPRITES', [])))
	SPRITES.clear()
	SPRITES.extend(sprites)

	assets = deepcopy(getattr(settings, 'ASSETS_MANAGER_FILES', {}))
	ASSETS.clear()
	ASSETS.update({n: v for n, v in assets.items()}) # cache is used in convert_asset_data
	ASSETS.update({n: convert_asset_data(n, v) for n, v in assets.items()})
	# download assets
	list(finder.list(ignore_patterns=[]))
	assets = deepcopy(getattr(settings, 'ASSETS_MANAGER_FILES', {}))
	ASSETS.update({n: v for n, v in assets.items()})
	ASSETS.update({n: convert_asset_data(n, v) for n, v in assets.items()})


update_settings()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import sys

import django
from django.conf import settings
from django.test.utils import get_runner


if __name__ == '__main__':
	os.environ['DJANGO_SETTINGS_MODULE'] = 'tests.settings'
	django.setup()
	TestRunner = get_runner(settings)
	test_runner = TestRunner()
	failures = test_runner.run_tests(sys.argv[1:])
	sys.exit(bool(failures))



# -*- coding: utf-8 -*-
from django.core.management.base import BaseCommand

from ... import utils
from ...settings import SPRITES


class Command(BaseCommand):
	requires_system_checks = []

	def handle(self, *args, **opions): #pylint: disable=unused-argument
		compiler = utils.SpriteCompiler()
		compiler.compile(SPRITES)



# -*- coding: utf-8 -*-
import os
import shutil
from datetime import datetime
from pathlib import Path

import httpretty
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.management import call_command
from django.template.loader import get_template
from django.test import TestCase, override_settings
from jinja2.runtime import Context

from PIL import Image
from django_assets_manager.checks import check_generated
from django_assets_manager.utils import NoSpaceError, AssetNotFoundError
from django_assets_manager.templatetags.assets_manager import assets, assets_by_type


def get_static_path(path: str) -> Path:
	static_dir = Path(settings.STATICFILES_DIRS[0])
	return Path.joinpath(static_dir, path)


def clear_cached_static_files():
	generated_dir = get_static_path('CACHE')
	if generated_dir.exists():
		shutil.rmtree(generated_dir)


class TemplateContextMixin(object):
	def ctx(self):
		return {}

	def jinja_ctx(self):
		tpl = get_template('index.html', using='jinja')
		environment = tpl.backend.env
		return Context(environment, parent={}, name='index.html', blocks={})


class TestAssets(TemplateContextMixin, TestCase):
	@override_settings(
		ASSETS_MANAGER_FILES = {
			'app': {
				'js': 'static://js/app.js',
				'css': 'static://js/app.css',
			},
		},
		X='x'
	)
	def test_single_asset(self):
		self.assertEqual('<link rel="stylesheet" href="/static/js/app.css" /><script src="/static/js/app.js"></script>', assets(self.ctx(), 'app'))

	@override_settings(
		ASSETS_MANAGER_FILES = {
			'app': {
				'js': 'static://js/app.js',
				'css': 'static://js/app.css',
			},
		},
	)
	def test_dont_repeat(self):
		ctx = self.ctx()
		assets(ctx, 'app')
		self.assertEqual('', assets(ctx, 'app'))

		ctx = self.jinja_ctx()
		assets(ctx, 'app')
		self.assertEqual('', assets(ctx, 'app'))

	@override_settings(
		ASSETS_MANAGER_FILES = {
			'app': {
				'js': 'http://example.tld/app.js',
				'css': 'http://example.tld/app.css',
			},
		},
	)
	def test_external(self):
		self.assertEqual('<link rel="stylesheet" href="http://example.tld/app.css" /><script src="http://example.tld/app.js"></script>', assets(self.ctx(), 'app'))

	@override_settings(
		ASSETS_MANAGER_FILES = {
			'app': {
				'js': ['static://1.js', 'static://2.js'],
				'css': ['static://1.css', 'static://2.css'],
			},
		},
	)
	def test_array(self):
		self.assertEqual('<link rel="stylesheet" href="/static/1.css" /><link rel="stylesheet" href="/static/2.css" /><script src="/static/1.js"></script><script src="/static/2.js"></script>', assets(self.ctx(), 'app'))

	@override_settings(
		ASSETS_MANAGER_FILES = {
		},
	)
	def test_throw_error(self):
		with self.assertRaises(ImproperlyConfigured):
			assets(self.ctx(), 'app')

	@override_settings(
		ASSETS_MANAGER_FILES = {
			'dep': {
				'js': ['static://1.js'],
			},
			'app': {
				'js': ['static://2.js'],
				'depends': ['dep'],
			},
		},
	)
	def test_dependencies(self):
		self.assertEqual('<script src="/static/1.js"></script><script src="/static/2.js"></script>', assets(self.ctx(), 'app'))

	@override_settings(
		ASSETS_MANAGER_FILES = {
			'app': {
				'custom': ['custom.script'],
			},
		},
	)
	def test_custom_type(self):
		self.assertEqual('Custom: custom.script', assets_by_type(self.ctx(), 'custom', 'app').strip())


class ImagesTestMixin(object):
	def create_image(self, path: str, width: int, height: int):
		image_path = get_static_path(path)
		image_path.parent.mkdir(exist_ok=True, parents=True)
		img = Image.new('RGB', (width, height))
		img.save(image_path)



class TestChecks(ImagesTestMixin, TestCase):
	def setUp(self):
		clear_cached_static_files()

	@classmethod
	def tearDownClass(cls):
		clear_cached_static_files()

	@override_settings(
		ASSETS_MANAGER_SPRITES = [
			{
				'name': 'main',
				'output': 'CACHE/sprites.png',
				'scss_output': 'CACHE/sprites.scss',
				'extra_sizes': ((2, '@2x'),),
				'width': 640,
				'height': 640,
				'images': (
					{
						'name': 'img',
						'src': 'img.png',
					},
				),
			},
		],
	)
	def test_recompilation_needed(self):
		# Create test images
		self.create_image('img.png', width=1, height=1)
		self.create_image('img@2x.png', width=2, height=2)

		# not generated
		errors = check_generated()
		self.assertEqual(1, len(errors))

		# now call generate sprites
		call_command('compilesprites')
		errors = check_generated()
		self.assertEqual(0, len(errors))

		# now pretend, that generated file is order
		older_time = int(datetime.now().timestamp()) - 1000000000
		generated_file = get_static_path('CACHE/sprites.png')
		os.utime(generated_file, (older_time, older_time))

		# check again
		errors = check_generated()
		self.assertEqual(1, len(errors))

	@override_settings(
		ASSETS_MANAGER_SPRITES = [
			{
				'name': 'main',
				'output': 'CACHE/sprites.png',
				'scss_output': 'CACHE/sprites.scss',
				'extra_sizes': [],
				'width': 640,
				'height': 640,
				'images': (
					{
						'name': 'notfound',
						'src': 'notfound.png',
					},
				),
			},
		],
	)
	def test_asset_not_found(self):
		with self.assertRaises(AssetNotFoundError):
			check_generated()


class TestCdnFinder(TemplateContextMixin, TestCase):
	def setUp(self):
		clear_cached_static_files()

	@classmethod
	def tearDownClass(cls):
		clear_cached_static_files()

	@override_settings(
		ASSETS_MANAGER_FILES = {
			'app': {
				'js': '//external.com/script.js',
			},
		},
	)
	def test_get_external_files(self):
		self.assertEqual('<script src="//external.com/script.js"></script>', assets(self.ctx(), 'app'))

	@httpretty.activate
	def test_cached_external_file(self):
		httpretty.register_uri(httpretty.GET, "http://external.com/script.js", body="")
		httpretty.register_uri(httpretty.GET, "http://external.com/script2.js", body="")
		with override_settings(
			ASSETS_MANAGER_FILES = {
				'app': {
					'js': 'http://external.com/script.js',
					'cache': {
						'paths': {
							'http://external.com/script.js': 'script.js',
						}
					},
				},
			},
		):
			self.assertEqual('<script src="/static/CACHE/app/script.js"></script>', assets(self.ctx(), 'app'))
		# same request without http prefix
		with override_settings(
			ASSETS_MANAGER_FILES = {
				'app': {
					'js': ['http://external.com/script.js', '//external.com/script2.js'],
					'cache': {
						'paths': {
							'http://external.com/script.js': 'script.js',
							'//external.com/script2.js': 'script2.js',
						}
					},
				},
			},
		):
			self.assertEqual('<script src="/static/CACHE/app/script.js"></script><script src="/static/CACHE/app/script2.js"></script>', assets(self.ctx(), 'app'))


class TestCompilesprites(TemplateContextMixin, ImagesTestMixin, TestCase):
	@override_settings(
		ASSETS_MANAGER_SPRITES = [
			{
				'name': 'main',
				'output': 'CACHE/sprites.png',
				'scss_output': 'CACHE/sprites.scss',
				'extra_sizes': [],
				'width': 1,
				'height': 1,
				'images': (
					{
						'name': 'src.png',
						'src': 'CACHE/src.png',
					},
				),
			},
		],
	)
	def test_single_sprite(self):
		self.create_image('CACHE/src.png', width=1, height=1)
		call_command('compilesprites')

	@override_settings(
		ASSETS_MANAGER_SPRITES = [
			{
				'name': 'main',
				'output': 'CACHE/sprites.png',
				'scss_output': 'CACHE/sprites.scss',
				'extra_sizes': [],
				'width': 1,
				'height': 1,
				'images': (
					{
						'name': 'src.png',
						'src': 'CACHE/src.png',
						'width': 1,
						'height': 1,
					},
				),
			},
		],
	)
	def test_defined_size(self):
		self.create_image('CACHE/src.png', width=1, height=1)
		call_command('compilesprites')

	@override_settings(
		ASSETS_MANAGER_SPRITES = [
			{
				'name': 'main',
				'output': 'CACHE/sprites.png',
				'scss_output': 'CACHE/sprites.scss',
				'extra_sizes': [],
				'width': 1,
				'height': 1,
				'images': (
					{
						'name': 'src.png',
						'src': 'CACHE/src.png',
						'mode': 'no-repeat',
					},
				),
			},
		],
	)
	def test_simple_no_space(self):
		self.create_image('CACHE/src.png', width=2, height=2)
		with self.assertRaises(NoSpaceError):
			call_command('compilesprites')

	@override_settings(
		ASSETS_MANAGER_SPRITES = [
			{
				'name': 'main',
				'output': 'CACHE/sprites.png',
				'scss_output': 'CACHE/sprites.scss',
				'extra_sizes': [],
				'width': 2,
				'height': 2,
				'images': (
					{'name': '0.png','src': 'CACHE/0.png'},
					{'name': '1.png','src': 'CACHE/1.png'},
					{'name': '2.png','src': 'CACHE/2.png'},
					{'name': '3.png','src': 'CACHE/3.png'},
				),
			},
		],
	)
	def test_needs_1px_space(self):
		for i in range(4):
			self.create_image(f'CACHE/{i}.png', width=1, height=1)
		with self.assertRaises(NoSpaceError):
			call_command('compilesprites')

	@override_settings(
		ASSETS_MANAGER_SPRITES = [
			{
				'name': 'main',
				'output': 'CACHE/sprites.png',
				'scss_output': 'CACHE/sprites.scss',
				'extra_sizes': [],
				'width': 3,
				'height': 3,
				'images': (
					{'name': '0.png','src': 'CACHE/0.png'},
					{'name': '1.png','src': 'CACHE/1.png'},
					{'name': '2.png','src': 'CACHE/2.png'},
					{'name': '3.png','src': 'CACHE/3.png'},
				),
			},
		],
	)
	def test_enough_space(self):
		for i in range(4):
			self.create_image(f'CACHE/{i}.png', width=1, height=1)
		call_command('compilesprites')

	@override_settings(
		ASSETS_MANAGER_SPRITES = [
			{
				'name': 'main',
				'output': 'CACHE/sprites.png',
				'scss_output': 'CACHE/sprites.scss',
				'extra_sizes': [],
				'width': 3,
				'height': 3,
				'images': (
					{'name': 'normal','src': 'CACHE/normal.png'},
					{'name': 'repeat','src': 'CACHE/repeat.png', 'mode': 'repeat-x'},
				),
			},
		],
	)
	def test_repeat_x(self):
		self.create_image(f'CACHE/normal.png', width=1, height=1)
		self.create_image(f'CACHE/repeat.png', width=1, height=1)
		call_command('compilesprites')

	@override_settings(
		ASSETS_MANAGER_SPRITES = [
			{
				'name': 'main',
				'output': 'CACHE/sprites.png',
				'scss_output': 'CACHE/sprites.scss',
				'extra_sizes': [],
				'width': 3,
				'height': 3,
				'images': (
					{'name': 'normal','src': 'CACHE/normal.png'},
					{'name': 'repeat','src': 'CACHE/repeat.png', 'mode': 'repeat-y'},
				),
			},
		],
	)
	def test_repeat_y(self):
		self.create_image(f'CACHE/normal.png', width=1, height=1)
		self.create_image(f'CACHE/repeat.png', width=1, height=1)
		call_command('compilesprites')

	@override_settings(
		ASSETS_MANAGER_SPRITES = [
			{
				'name': 'main',
				'output': 'CACHE/sprites.png',
				'scss_output': 'CACHE/sprites.scss',
				'extra_sizes': [],
				'width': 3,
				'height': 3,
				'images': (
					{'name': 'normal','src': 'CACHE/normal.png', 'mode': 'repeat-y'},
					{'name': 'repeat','src': 'CACHE/repeat.png', 'mode': 'repeat-x'},
				),
			},
		],
	)
	def test_repeat_x_combined(self):
		self.create_image(f'CACHE/normal.png', width=2, height=2)
		self.create_image(f'CACHE/repeat.png', width=1, height=1)
		with self.assertRaises(NoSpaceError):
			call_command('compilesprites')

	@override_settings(
		ASSETS_MANAGER_SPRITES = [
			{
				'name': 'main',
				'output': 'CACHE/sprites.png',
				'scss_output': 'CACHE/sprites.scss',
				'extra_sizes': [],
				'width': 3,
				'height': 3,
				'images': (
					{'name': 'normal','src': 'CACHE/normal.png', 'mode': 'repeat-x'},
					{'name': 'repeat','src': 'CACHE/repeat.png', 'mode': 'repeat-y'},
				),
			},
		],
	)
	def test_repeat_y_combined(self):
		self.create_image(f'CACHE/normal.png', width=2, height=2)
		self.create_image(f'CACHE/repeat.png', width=1, height=1)
		with self.assertRaises(NoSpaceError):
			call_command('compilesprites')

	@override_settings(
		ASSETS_MANAGER_SPRITES = [
			{
				'name': 'main',
				'output': 'CACHE/sprites.png',
				'scss_output': 'CACHE/sprites.scss',
				'extra_sizes': [],
				'width': 1,
				'height': 1,
				'images': (
					{'name': 'notfound','src': 'CACHE/notfound.png'},
				),
			},
		],
	)
	def test_not_found(self):
		with self.assertRaises(AssetNotFoundError):
			call_command('compilesprites')

	@override_settings(
		ASSETS_MANAGER_SPRITES = [
			{
				'name': 'main',
				'output': 'CACHE/sprites.png',
				'scss_output': 'CACHE/sprites.scss',
				'extra_sizes': [],
				'width': 2,
				'height': 2,
				'images': (
					{'name': '1','src': 'CACHE/1.png', 'mode': 'repeat-x'},
					{'name': '2','src': 'CACHE/2.png', 'mode': 'repeat-x'},
				),
			},
		],
	)
	def test_repeat_x_no_space(self):
		self.create_image(f'CACHE/1.png', width=1, height=1)
		self.create_image(f'CACHE/2.png', width=1, height=1)
		with self.assertRaises(NoSpaceError):
			call_command('compilesprites')

	@override_settings(
		ASSETS_MANAGER_SPRITES = [
			{
				'name': 'main',
				'output': 'CACHE/sprites.png',
				'scss_output': 'CACHE/sprites.scss',
				'extra_sizes': [],
				'width': 2,
				'height': 2,
				'images': (
					{'name': '1','src': 'CACHE/1.png', 'mode': 'repeat-y'},
					{'name': '2','src': 'CACHE/2.png', 'mode': 'repeat-y'},
				),
			},
		],
	)
	def test_repeat_y_no_space(self):
		self.create_image(f'CACHE/1.png', width=1, height=1)
		self.create_image(f'CACHE/2.png', width=1, height=1)
		with self.assertRaises(NoSpaceError):
			call_command('compilesprites')



# -*- coding: utf-8 -*-
from django.apps import AppConfig
from django.core import checks

from .checks import check_generated


class AssetsManagerConfig(AppConfig):
	name = 'django_assets_manager'
	verbose_name = "Assets manager"

	def ready(self):
		checks.register()(check_generated)




# -*- coding: utf-8 -*-
from django.core.checks import Error

from .settings import SPRITES
from .utils import SpriteCompiler


def check_generated(**kwargs):
	errors = []

	compiler = SpriteCompiler()
	if compiler.recompilation_needed(SPRITES):
		errors.append(Error(
			'Sprites not generated',
			hint='Run manage.py compilesprites',
			id='django_assets_manager.E001'
		))
	return errors



# -*- coding: utf-8 -*-
import traceback
from itertools import chain

from django import template
from django.core.exceptions import ImproperlyConfigured
from django.template.loader import render_to_string
from django.utils.safestring import mark_safe

from ..settings import ASSETS, USE_TEMPLATES


register = template.Library()


class ShadowRequest(object):
	pass


class FakeRequest(object):
	def __init__(self, *args, **kwargs):
		super(FakeRequest, self).__init__(*args, **kwargs)
		self.__dict__['requests'] = [ShadowRequest(), ShadowRequest()]

	def __getattr__(self, item):
		fake = any(f[2] == 'render_nodelist' for f in traceback.extract_stack())
		return getattr(self.__dict__['requests'][0 if fake else 1], item)

	def __setattr__(self, item, value):
		fake = any(f[2] == 'render_nodelist' for f in traceback.extract_stack())
		return setattr(self.__dict__['requests'][0 if fake else 1], item, value)


def render_css(context):
	return ''.join(f'<link rel="stylesheet" href="{css}" />' for css, __ in context['data'])


def render_js(context):
	return ''.join(f'<script src="{src}"{attributes}></script>' for src, attributes in context['data'])


def get_asset_sources(asset, unused, asset_type, render):
	if not asset in unused:
		if not asset in ASSETS:
			raise ImproperlyConfigured("Asset %s not registered" % asset)
		return ()

	asset_data = ASSETS[asset]
	unused.remove(asset)

	sources = []
	for depend in asset_data["depends"]:
		sources += get_asset_sources(depend, unused, asset_type, render)

	if asset_data[asset_type]:
		sources.append(render({'data': asset_data[asset_type]}))
	return sources


def get_or_create_unused_assets(context, asset_type):
	if not "request" in context:
		extra = {'request': FakeRequest()}
		if hasattr(context, 'update'):
			context.update(extra)
		else:
			context.vars.update(extra)
	if hasattr(context["request"], "unused_" + asset_type):
		return getattr(context["request"], "unused_" + asset_type)
	else:
		unused_assets = set(ASSETS.keys())
		setattr(context["request"], "unused_" + asset_type, unused_assets)
		return unused_assets


def assets_by_type(context, asset_type, *asset_list):
	unused = get_or_create_unused_assets(context, asset_type)
	if USE_TEMPLATES or asset_type not in ('css', 'js'):
		render = lambda context: render_to_string("assets_manager/" + asset_type + ".html", context)
	else:
		render = render_css if asset_type == 'css' else render_js

	return ''.join(chain(*(get_asset_sources(asset, unused, asset_type, render) for asset in asset_list)))


@register.simple_tag(takes_context=True)
def assets_js(context, *asset_list):
	return mark_safe(assets_by_type(context, "js", *asset_list))


@register.simple_tag(takes_context=True)
def assets_css(context, *asset_list):
	return mark_safe(assets_by_type(context, "css", *asset_list))


@register.simple_tag(takes_context=True)
def assets(context, *asset_list):
	return mark_safe(assets_css(context, *asset_list) + assets_js(context, *asset_list))


try: # pragma: no cover
	from django_jinja import library
	try:
		from jinja2 import pass_context
	except ImportError:
		from jinja2 import contextfunction as pass_context

	library.global_function(pass_context(assets_js))
	library.global_function(pass_context(assets_css))
	library.global_function(pass_context(assets))
except ImportError: # pragma: no cover
	pass



# -*- coding: utf-8 -*-
from pathlib import Path

BASE_DIR = Path(__file__).parent

INSTALLED_APPS = ['tests', 'django_assets_manager']
SECRET_KEY = 'secret'
DEFAULT_AUTO_FIELD = 'django.db.models.BigAutoField'
ROOT_URLCONF = 'tests.urls'
USE_TZ = False
STATICFILES_DIRS = [BASE_DIR / 'static']
STATIC_URL = '/static/'

DATABASES = {
	'default': {
		'ENGINE': 'django.db.backends.sqlite3',
		'NAME': ':memory:',
	}
}

STATICFILES_FINDERS = (
	'django.contrib.staticfiles.finders.FileSystemFinder',
	'django.contrib.staticfiles.finders.AppDirectoriesFinder',
	'django_assets_manager.finders.CdnFinder',
)
TEMPLATES = [
	{
		'BACKEND': 'django_jinja.backend.Jinja2',
		'NAME': 'jinja',
		'DIRS': [BASE_DIR / 'templates' / 'jinja'],
		'APP_DIRS': False,
		'OPTIONS': {
			"match_extension": None,
		}
	},
	{
		'BACKEND': 'django.template.backends.django.DjangoTemplates',
		'NAME': 'django',
		'DIRS': [BASE_DIR / 'templates' / 'django'],
		'APP_DIRS': False,
		'OPTIONS': {
			'context_processors': ['django.template.context_processors.request'],
			'builtins': ['django_assets_manager.templatetags.assets_manager'],
		},
	},
]



# -*- coding: utf-8 -*-
import os
from pathlib import Path

from django.conf import settings as django_settings
from django.contrib.staticfiles.finders import BaseFinder
from django.contrib.staticfiles.utils import matches_patterns
from django.core.files.storage import FileSystemStorage
from urllib.request import urlopen

from . import settings


class CdnFinder(BaseFinder):
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.storage = FileSystemStorage(location=django_settings.STATICFILES_DIRS[0])

	def find(self, path, all=False): #pylint: disable=redefined-builtin
		return [] # pragma: no cover

	def list(self, ignore_patterns):
		for package, data in settings.ASSETS.items():
			if "cache" in data:
				for src, dest in data["cache"]["paths"].items():
					dest_path = Path(self.storage.path(self.to_cache_path(package, dest)))
					if matches_patterns(dest, ignore_patterns): # pragma: no cover
						continue
					if not dest_path.exists():
						dest_path.parent.mkdir(parents=True, exist_ok=True)
						content = urlopen(self.to_url(src)).read()
						with dest_path.open('wb') as fp:
							fp.write(content)
					yield dest, self.storage

	def to_cache_path(self, package, dest):
		return os.path.join('CACHE', package, dest)

	def transform_and_check_path(self, package, path):
		dest_cache = settings.ASSETS.get(package, {}).get('cache', {}).get('paths', {})
		if not path in dest_cache:
			return path
		dest = self.to_cache_path(package, dest_cache[path])
		cache_path = Path(self.storage.path(dest))
		if cache_path.exists():
			return django_settings.STATIC_URL + dest
		else:
			return path

	def transform_to_cache(self, package, paths):
		return [self.transform_and_check_path(package, path) for path in paths]

	def to_url(self, url):
		if url[:2] == '//':
			url = 'http:' + url
		return url



# -*- coding: utf-8 -*-
import os
from copy import deepcopy
from pathlib import Path

from django.conf import settings
from django.contrib.staticfiles import finders


def find_file(path):
	return finders.find(path)


def to_localfile(path) -> Path:
	return Path.joinpath(Path(settings.STATICFILES_DIRS[0]), path)


class NoSpaceError(RuntimeError):
	pass


class AssetNotFoundError(RuntimeError):
	pass


class Packer:
	def __init__(self, width, height):
		self.root = {
			'pos': (0, 0),
			'size': (width + 1, height + 1), # 1px gap
			'used': False,
			'down': None,
			'right': None
		}

	def fit(self, blocks):
		repeat_mode = None

		for block in blocks:
			if not 'mode' in block:
				block['mode'] = 'no-repeat'
			if block['mode'] == 'repeat-x':
				self.fit_block_repeat_x(block)
				if repeat_mode is None or repeat_mode == 'x':
					repeat_mode = 'x'
				else:
					raise NoSpaceError('Can not mix repeat-x and repeat-y for %s' % block['name'])
			elif block['mode'] == 'repeat-y':
				self.fit_block_repeat_y(block)
				if repeat_mode is None or repeat_mode == 'y':
					repeat_mode = 'y'
				else:
					raise NoSpaceError('Can not mix repeat-x and repeat-y for %s' % block['name'])

		for block in blocks:
			if block['mode'] == 'no-repeat':
				self.fit_block(block)

	def fit_block(self, block):
		size = (block['width'] + 1, block['height'] + 1) # 1px gap
		node = self.find_node(self.root, size)
		if node is None:
			raise NoSpaceError('Block %s' % block['name'])
		self.split_node(node, size)
		node['used'] = block['name']
		block['pos'] = node['pos']
		block['size'] = (block['width'], block['height'])

	def fit_block_repeat_x(self, block):
		self.root['size'] = (self.root['size'][0], self.root['size'][1] - block['height'] - 1)
		if self.root['size'][1] <= -1:
			raise NoSpaceError('Block %s' % block['name'])

		block['pos'] = (0, self.root['size'][1])
		block['size'] = (self.root['size'][0] - 1, block['height'])

	def fit_block_repeat_y(self, block):
		self.root['size'] = (self.root['size'][0] - block['width'] - 1, self.root['size'][1])
		if self.root['size'][0] <= -1:
			raise NoSpaceError('Block %s' % block['name'])

		block['pos'] = (self.root['size'][0], 0)
		block['size'] = (block['width'], self.root['size'][1] - 1)

	def find_node(self, root, size):
		#if not root:
		#	return None

		if root['used']:
			return self.find_node(root['right'], size) or self.find_node(root['down'], size)

		w, h = size
		root_w, root_h = root['size']
		if w <= root_w and h <= root_h:
			return root

		return None

	def split_node(self, root, size):
		root_w, root_h = root['size']
		w, h = size
		x, y = root['pos']

		root['down'] =  {'pos': (x, y + h), 'size': (root_w, root_h - h), 'used': False, 'down': None, 'right': None}
		root['right'] = {'pos': (x + w, y), 'size': (root_w - w, h), 'used': False, 'down': None, 'right': None}


class SpriteGenerator:
	def __init__(self, filename, size, pixel_ratio):
		from PIL import Image
		self.filename = filename
		self.size = size
		self.pixel_ratio = pixel_ratio
		self.out_image = Image.new('RGBA', (size[0] * pixel_ratio, size[1] * pixel_ratio))

	def generate(self, images):
		for img in images:
			self.paste_image(img)
		output_filename = to_localfile(self.filename)
		output_filename.parent.mkdir(parents=True, exist_ok=True)
		self.out_image.save(output_filename)

	def paste_image(self, image):
		from PIL import Image
		in_image = Image.open(find_file(image['src']))
		if image['mode'] == 'repeat-x':
			w = image['width']
			for i in range((self.size[0] + w - 1) // w):
				self.out_image.paste(in_image, (i * w * self.pixel_ratio, image['pos'][1] * self.pixel_ratio))
		elif image['mode'] == 'repeat-y':
			h = image['height']
			for i in range((self.size[1] + h - 1) // h):
				self.out_image.paste(in_image, (image['pos'][0] * self.pixel_ratio, i * h * self.pixel_ratio))
		else:
			self.out_image.paste(in_image, (image['pos'][0] * self.pixel_ratio, image['pos'][1] * self.pixel_ratio))

	def generate_scss(self, sprites, sprite_configs):
		metadata = {
			'_w': str(sprites['width']) + 'px',
			'_h': str(sprites['height']) + 'px',
		}
		metadata['_size'] = metadata['_w'] + ' ' + metadata['_h']
		metadata['_ratio'] = '(' + (' '.join(str(config['ratio']) for config in sprite_configs)) + ')'
		metadata['_url'] = '(' + (' '.join('url(static("' + config['output'] + '"))' for config in sprite_configs)) + ')'

		f = open(to_localfile(sprites['scss_output']), 'w+')
		f.write('$' + sprites['name'] + ': (\n')
		for k, v in metadata.items():
			f.write(k + ': ' + v + ',\n')
		f.write(',\n'.join([self.generate_image_scss(img) for img in sprite_configs[0]['images']]))
		f.write('\n);')

	def generate_image_scss(self, image):
		w, h = (str(image['width']) + 'px', str(image['height']) + 'px')
		x, y = (str(image['pos'][0]) + 'px', str(image['pos'][1]) + 'px')
		ctx = {
			'name': image['original'],
			'w': w,
			'h': h,
			'x': x,
			'y': y,
			'size': w + ' ' + h,
			'offset': '-' + x + ' -' + y,
		}
		return '{name}: (w: {w}, h: {h}, x: {x}, y: {y}, size: {size}, offset: {offset})'.format(**ctx)


class SpriteCompiler:
	def compile(self, sprites):
		for sprite_def in sprites:
			self.pack_sprites(sprite_def)

	def pack_sprites(self, sprites):
		from PIL import Image
		sizes = ((1, ''),)
		sizes += tuple(sprites.get('extra_sizes', ()))

		for img in sprites['images']:
			if not 'width' in img or not 'height' in img:
				src_filename = find_file(img['src'])
				if src_filename is None:
					raise AssetNotFoundError("File %s not found" % img['src'])
				(width, height) = Image.open(src_filename).size
				img['width'] = width
				img['height'] = height

		sprite_configs = []
		for size, suffix in sizes:
			sprite_conf = self.preprocess_pixel_ratio(sprites, (size, suffix))
			packer = Packer(sprite_conf['width'], sprite_conf['height'])
			packer.fit(sprite_conf['images'])
			generator = SpriteGenerator(sprite_conf['output'], (sprite_conf['width'], sprite_conf['height']), size)
			generator.generate(sprite_conf['images'])
			sprite_configs.append(sprite_conf)
		generator.generate_scss(sprites, sprite_configs)

	def add_suffix(self, name, suffix):
		return suffix.join(os.path.splitext(name))

	def preprocess_pixel_ratio(self, sprites, size):
		ratio, suffix = size
		sprites = deepcopy(sprites)
		del sprites['extra_sizes']
		sprites['output'] = self.add_suffix(sprites['output'], suffix)
		sprites['ratio'] = ratio
		sprites['suffix'] = suffix

		for image in sprites['images']:
			image['original'] = image['name']
			image['name'] = image['name'] + suffix
			image['src'] = self.add_suffix(image['src'], suffix)

		return sprites

	def get_mtime(self, filename):
		try:
			return os.path.getmtime(filename)
		except OSError:
			return None

	def get_dependencies(self, sprites):
		dependencies = {}

		for sprite_def in sprites:
			sizes = ((1, ''),)
			sizes += tuple(sprite_def.get('extra_sizes', ()))

			for size, suffix in sizes:
				sprite_conf = self.preprocess_pixel_ratio(sprite_def, (size, suffix))
				deps = []
				dependencies[sprite_conf['output']] = {
					'ts': self.get_mtime(to_localfile(sprite_conf['output'])),
					'dep': deps,
				}
				for img in sprite_conf['images']:
					src_filename = find_file(img['src'])
					if src_filename is None:
						raise AssetNotFoundError("File %s not found" % img['src'])
					deps.append((img['src'], self.get_mtime(find_file(img['src']))))

		return dependencies

	def recompilation_needed(self, sprites):
		deps = self.get_dependencies(sprites)
		for dep in deps.values():
			if dep['ts'] is None:
				return True
			if any(img[1] > dep['ts'] for img in dep['dep']):
				return True
		return False
